# -*- coding: utf-8 -*-
"""Redis store.

Only supports task storage at the moment.
"""

import uuid

import redis  # pylint: disable=import-error

from acstore import interface
from acstore.containers import interface as containers_interface

from plaso.containers import events
from plaso.lib import definitions
from plaso.serializer import json_serializer
from plaso.storage import logger


class RedisStore(interface.AttributeContainerStore):
  """Redis store.

  Attribute containers are stored as Redis Hashes.
  All keys are prefixed with the session identifier to avoid collisions.
  Event identifiers are also stored in an index to enable sorting.
  """

  _CONTAINER_TYPE_EVENT = events.EventObject.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA = events.EventData.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA_STREAM = events.EventDataStream.CONTAINER_TYPE

  _FORMAT_VERSION = '20181013'
  _EVENT_INDEX_NAME = 'sorted_event_identifier'

  DEFAULT_REDIS_URL = 'redis://127.0.0.1/0'

  def __init__(self):
    """Initializes a Redis store."""
    super(RedisStore, self).__init__()
    self._redis_client = None
    self._session_identifier = None
    self._serializer = json_serializer.JSONAttributeContainerSerializer
    self._serializers_profiler = None
    self._task_identifier = None

    self.serialization_format = definitions.SERIALIZER_FORMAT_JSON

  def _DeserializeAttributeContainer(self, container_type, serialized_data):
    """Deserializes an attribute container.

    Args:
      container_type (str): attribute container type.
      serialized_data (bytes): serialized attribute container data.

    Returns:
      AttributeContainer: attribute container or None.

    Raises:
      IOError: if the serialized data cannot be decoded.
      OSError: if the serialized data cannot be decoded.
    """
    if not serialized_data:
      return None

    if self._serializers_profiler:
      self._serializers_profiler.StartTiming(container_type)

    try:
      serialized_string = serialized_data.decode('utf-8')
      attribute_container = self._serializer.ReadSerialized(serialized_string)

    except UnicodeDecodeError as exception:
      raise IOError('Unable to decode serialized data: {0!s}'.format(exception))

    except (TypeError, ValueError) as exception:
      # TODO: consider re-reading attribute container with error correction.
      raise IOError('Unable to read serialized data: {0!s}'.format(exception))

    finally:
      if self._serializers_profiler:
        self._serializers_profiler.StopTiming(container_type)

    return attribute_container

  def _GetRedisHashName(self, container_type):
    """Retrieves the Redis hash name of the attribute container type.

    Args:
      container_type (str): container type.

    Returns:
      str: a Redis key name.
    """
    return '{0:s}-{1:s}-{2:s}'.format(
        self._session_identifier, self._task_identifier, container_type)

  def _RaiseIfNotReadable(self):
    """Checks that the store is ready to for reading.

     Raises:
       IOError: if the store cannot be read from.
       OSError: if the store cannot be read from.
    """
    if not self._redis_client:
      raise IOError('Unable to read, client not connected.')

  def _RaiseIfNotWritable(self):
    """Checks that the store is ready to for writing.

    Raises:
      IOError: if the store cannot be written to.
      OSError: if the store cannot be written to.
    """
    if not self._redis_client:
      raise IOError('Unable to write, client not connected.')

  def _SerializeAttributeContainer(self, attribute_container):
    """Serializes an attribute container.

    Args:
      attribute_container (AttributeContainer): attribute container.

    Returns:
      bytes: serialized attribute container.

    Raises:
      IOError: if the attribute container cannot be serialized.
      OSError: if the attribute container cannot be serialized.
    """
    if self._serializers_profiler:
      self._serializers_profiler.StartTiming(
          attribute_container.CONTAINER_TYPE)

    try:
      attribute_container_data = self._serializer.WriteSerialized(
          attribute_container)
      if not attribute_container_data:
        raise IOError(
            'Unable to serialize attribute container: {0:s}.'.format(
                attribute_container.CONTAINER_TYPE))

    finally:
      if self._serializers_profiler:
        self._serializers_profiler.StopTiming(
            attribute_container.CONTAINER_TYPE)

    return attribute_container_data

  @classmethod
  def _SetClientName(cls, redis_client, name):
    """Attempts to sets a Redis client name.

    This method ignores errors from the Redis server or exceptions
    indicating the method is missing, as setting the name is not a critical
    function, and it is not currently supported by the fakeredis test library.

    Args:
      redis_client (Redis): an open Redis client.
      name (str): name to set.
    """
    try:
      redis_client.client_setname(name)
    except redis.ResponseError as exception:
      logger.debug(
          'Unable to set redis client name: {0:s} with error: {1!s}'.format(
              name, exception))

  def _UpdateAttributeContainerAfterDeserialize(self, container):
    """Updates an attribute container after deserialization.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      ValueError: if an attribute container identifier is missing.
    """
    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT:
      identifier = getattr(container, '_event_data_identifier', None)
      if identifier:
        event_data_identifier = (
            containers_interface.AttributeContainerIdentifier(
                name=self._CONTAINER_TYPE_EVENT_DATA,
                sequence_number=identifier))
        container.SetEventDataIdentifier(event_data_identifier)

    elif container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
      identifier = getattr(container, '_event_data_stream_identifier', None)
      if identifier:
        event_data_stream_identifier = (
            containers_interface.AttributeContainerIdentifier(
                name=self._CONTAINER_TYPE_EVENT_DATA_STREAM,
                sequence_number=identifier))
        container.SetEventDataStreamIdentifier(event_data_stream_identifier)

  def _UpdateAttributeContainerBeforeSerialize(self, container):
    """Updates an attribute container before serialization.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      IOError: if the attribute container identifier type is not supported.
      OSError: if the attribute container identifier type is not supported.
    """
    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT:
      event_data_identifier = container.GetEventDataIdentifier()
      if event_data_identifier:
        setattr(container, '_event_data_identifier',
                event_data_identifier.sequence_number)

    elif container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
      event_data_stream_identifier = container.GetEventDataStreamIdentifier()
      if event_data_stream_identifier:
        setattr(container, '_event_data_stream_identifier',
                event_data_stream_identifier.sequence_number)

  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      IOError: if an unsupported identifier is provided or if the attribute
          container does not exist.
      RuntimeError: since this method is not implemented.
      OSError: if an unsupported identifier is provided or if the attribute
          container does not exist.
    """
    identifier = container.GetIdentifier()
    redis_hash_name = self._GetRedisHashName(container.CONTAINER_TYPE)
    redis_key = identifier.CopyToString()

    self._UpdateAttributeContainerBeforeSerialize(container)

    serialized_data = self._SerializeAttributeContainer(container)
    self._redis_client.hset(
        redis_hash_name, key=redis_key, value=serialized_data)

  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """
    next_sequence_number = self._GetAttributeContainerNextSequenceNumber(
        container.CONTAINER_TYPE)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container.CONTAINER_TYPE, sequence_number=next_sequence_number)
    container.SetIdentifier(identifier)

    redis_hash_name = self._GetRedisHashName(container.CONTAINER_TYPE)
    redis_key = identifier.CopyToString()

    self._UpdateAttributeContainerBeforeSerialize(container)

    serialized_data = self._SerializeAttributeContainer(container)
    self._redis_client.hsetnx(redis_hash_name, redis_key, serialized_data)

    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT:
      index_name = self._GetRedisHashName(self._EVENT_INDEX_NAME)
      self._redis_client.zincrby(index_name, container.timestamp, redis_key)

  def _WriteStorageMetadata(self):
    """Writes the storage metadata."""
    metadata = {
        'format_version': self._FORMAT_VERSION,
        'serialization_format': self.serialization_format}
    metadata_key = self._GetRedisHashName('metadata')

    for key, value in metadata.items():
      self._redis_client.hset(metadata_key, key=key, value=value)

  def Close(self):
    """Closes the store.

    Raises:
      IOError: if the store is already closed.
      OSError: if the store is already closed.
    """
    if not self._redis_client:
      raise IOError('Store already closed.')

    self._redis_client = None

  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when the store is closed or if an unsupported identifier is
          provided.
      OSError: when the store is closed or if an unsupported identifier is
          provided.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    redis_key = identifier.CopyToString()

    serialized_data = self._redis_client.hget(redis_hash_name, redis_key)
    if not serialized_data:
      return None

    attribute_container = self._DeserializeAttributeContainer(
        container_type, serialized_data)

    attribute_container.SetIdentifier(identifier)

    self._UpdateAttributeContainerAfterDeserialize(attribute_container)

    return attribute_container

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    sequence_number = index + 1
    redis_hash_name = self._GetRedisHashName(container_type)
    redis_key = '{0:s}.{1:d}'.format(container_type, sequence_number)

    serialized_data = self._redis_client.hget(redis_hash_name, redis_key)
    if not serialized_data:
      return None

    attribute_container = self._DeserializeAttributeContainer(
        container_type, serialized_data)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container_type, sequence_number=sequence_number)
    attribute_container.SetIdentifier(identifier)

    self._UpdateAttributeContainerAfterDeserialize(attribute_container)

    return attribute_container

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves attribute containers

    Args:
      container_type (str): container type attribute of the container being
          added.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Yields:
      AttributeContainer: attribute container.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    for redis_key, serialized_data in self._redis_client.hscan_iter(
        redis_hash_name):
      redis_key = redis_key.decode('utf-8')

      attribute_container = self._DeserializeAttributeContainer(
          container_type, serialized_data)

      _, sequence_number = redis_key.split('.')
      sequence_number = int(sequence_number, 10)
      identifier = containers_interface.AttributeContainerIdentifier(
          name=container_type, sequence_number=sequence_number)
      attribute_container.SetIdentifier(identifier)

      self._UpdateAttributeContainerAfterDeserialize(attribute_container)

      # TODO: map filter expression to Redis native filter.
      if attribute_container.MatchesExpression(filter_expression):
        yield attribute_container

  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    return self._redis_client.hlen(redis_hash_name)

  def GetSerializedAttributeContainers(
      self, container_type, cursor, maximum_number_of_items):
    """Fetches serialized attribute containers.

    Args:
      container_type (str): attribute container type.
      cursor (int): Redis cursor.
      maximum_number_of_items (int): maximum number of containers to
          retrieve, where 0 represent no limit.

    Returns:
      tuple: containing:
        int: Redis cursor.
        list[bytes]: serialized attribute containers.
    """
    name = self._GetRedisHashName(container_type)
    # Redis treats None as meaning "no limit", not 0.
    if maximum_number_of_items == 0:
      maximum_number_of_items = None

    cursor, items = self._redis_client.hscan(
        name, cursor=cursor, count=maximum_number_of_items)
    return cursor, items

  def GetSortedEvents(self, time_range=None):
    """Retrieves the events in increasing chronological order.

    Args:
      time_range (Optional[TimeRange]): This argument is not supported by the
          Redis store.

    Yields:
      EventObject: event.

    Raises:
      RuntimeError: if a time_range argument is specified.
    """
    event_index_name = self._GetRedisHashName(self._EVENT_INDEX_NAME)
    if time_range:
      raise RuntimeError('Not supported')

    for redis_key, _ in self._redis_client.zscan_iter(event_index_name):
      redis_key = redis_key.decode('utf-8')

      container_type, sequence_number = redis_key.split('.')
      sequence_number = int(sequence_number, 10)
      identifier = containers_interface.AttributeContainerIdentifier(
          name=container_type, sequence_number=sequence_number)
      yield self.GetAttributeContainerByIdentifier(
          self._CONTAINER_TYPE_EVENT, identifier)

  def HasAttributeContainers(self, container_type):
    """Determines if the store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    number_of_containers = self._redis_client.hlen(redis_hash_name)
    return number_of_containers > 0

  # pylint: disable=arguments-differ
  def Open(
      self, redis_client=None, session_identifier=None, task_identifier=None,
      url=None, **unused_kwargs):
    """Opens the store.

    Args:
      redis_client (Optional[Redis]): Redis client to query. If specified, no
          new client will be created. If no client is specified a new client
          will be opened connected to the Redis instance specified by 'url'.
      session_identifier (Optional[str]): identifier of the session.
      task_identifier (Optional[str]): unique identifier of the task the store
          will store containers for. If not specified, an identifier will be
          generated.
      url (Optional[str]): URL for a Redis database. If not specified, the
          DEFAULT_REDIS_URL will be used.

    Raises:
      IOError: if the store is already connected to a Redis instance.
      OSError: if the store is already connected to a Redis instance.
    """
    if self._redis_client:
      raise IOError('Redis client already connected')

    if not redis_client:
      if not url:
        url = self.DEFAULT_REDIS_URL

      redis_client = redis.from_url(url=url, socket_timeout=60)

    self._redis_client = redis_client

    self._session_identifier = session_identifier or str(uuid.uuid4())
    self._task_identifier = task_identifier or str(uuid.uuid4())

    client_name = self._GetRedisHashName('')
    self._SetClientName(self._redis_client, client_name)

    metadata_key = self._GetRedisHashName('metadata')
    if not self._redis_client.exists(metadata_key):
      self._WriteStorageMetadata()

  def SetSerializersProfiler(self, serializers_profiler):
    """Sets the serializers profiler.

    Args:
      serializers_profiler (SerializersProfiler): serializers profiler.
    """
    self._serializers_profiler = serializers_profiler



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Safari history database plugin."""

import unittest

from plaso.parsers.sqlite_plugins import safari

from tests.parsers.sqlite_plugins import test_lib


class SafariHistoryPluginTest(test_lib.SQLitePluginTestCase):
  """Tests for the Safari History database plugin."""

  def testProcess(self):
    """Tests the process function on a Safari History.db database file."""
    plugin = safari.SafariHistoryPluginSqlite()
    storage_writer = self._ParseDatabaseFileWithPlugin(['History.db'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 25)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'safari:history:visit_sqlite',
        'last_visited_time': '2017-11-09T21:24:28.829571+00:00',
        'title': None,
        'url': 'http://facebook.com/',
        'visit_count': 2,
        'was_http_non_get': False}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Windows known folders helper."""

import unittest

from plaso.helpers.windows import known_folders

from tests import test_lib as shared_test_lib


class WindowsKnownFoldersHelperTest(shared_test_lib.BaseTestCase):
  """Tests for the Windows known folders helper."""

  def testGetPath(self):
    """Tests the GetPath function."""
    path = known_folders.WindowsKnownFoldersHelper.GetPath(
        '{b4bfcc3a-db2c-424c-b029-7fe99a87c641}')
    self.assertEqual(path, '%USERPROFILE%\\Desktop')

    path = known_folders.WindowsKnownFoldersHelper.GetPath('bogus')
    self.assertIsNone(path)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""This file contains the interface for text plugins."""

import abc
import codecs
import os

import pyparsing

from plaso.lib import errors
from plaso.parsers import logger
from plaso.parsers import plugins
from plaso.parsers import text_parser


class TextPlugin(plugins.BasePlugin):
  """The interface for text plugins."""

  NAME = 'text_plugin'
  DATA_FORMAT = 'Text file'

  ENCODING = None

  # List of tuples of pyparsing expression per unique identifier that define
  # the supported grammar.
  _LINE_STRUCTURES = []

  # PyParsing grammer used to verify the text-log file format. Note that since
  # this is called often it should optimize on failing fast.
  VERIFICATION_GRAMMAR = None
  VERIFICATION_LITERALS = None

  # The maximum number of consecutive lines that do not match the grammar before
  # aborting parsing.
  _MAXIMUM_CONSECUTIVE_LINE_FAILURES = 20

  def __init__(self):
    """Initializes a parser."""
    super(TextPlugin, self).__init__()
    self._current_offset = 0
    self._parser_mediator = None
    self._pyparsing_grammar = None

    codecs.register_error('text_parser_handler', self._EncodingErrorHandler)

    self._SetLineStructures(self._LINE_STRUCTURES)

  def _EncodingErrorHandler(self, exception):
    """Encoding error handler.

    Args:
      exception [UnicodeDecodeError]: exception.

    Returns:
      tuple[str, int]: replacement string and a position where encoding should
          continue.

    Raises:
      TypeError: if exception is not of type UnicodeDecodeError.
    """
    if not isinstance(exception, UnicodeDecodeError):
      raise TypeError('Unsupported exception type.')

    if self._parser_mediator:
      self._parser_mediator.ProduceExtractionWarning(
          'error decoding 0x{0:02x} at offset: {1:d}'.format(
              exception.object[exception.start],
              self._current_offset + exception.start))

    escaped = '\\x{0:2x}'.format(exception.object[exception.start])
    return escaped, exception.start + 1

  def _GetDecimalValueFromStructure(self, structure, name):
    """Retrieves a decimal integer value from a Pyparsing structure.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.

    Returns:
      int: decimal integer value or None if not available or invalid.
    """
    integer_value = self._GetValueFromStructure(structure, name)

    try:
      return int(integer_value, 10)
    except (TypeError, ValueError):
      pass

    return None

  def _GetStringValueFromStructure(self, structure, name):
    """Retrieves a string value from a Pyparsing structure.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.

    Returns:
      str: string value or None if not available or empty.
    """
    string_value = self._GetValueFromStructure(
        structure, name, default_value='')
    return string_value.strip() or None

  def _GetValueFromStructure(self, structure, name, default_value=None):
    """Retrieves a token value from a Pyparsing structure.

    This method ensures the token value is set to the default value when
    the token is not present in the structure. Instead of returning
    the Pyparsing default value of an empty byte stream (b'').

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.
      default_value (Optional[object]): default value.

    Returns:
      object: value in the token or default value if the token is not available
          in the structure.
    """
    value = structure.get(name, default_value)
    if isinstance(value, pyparsing.ParseResults) and not value:
      # Ensure the return value is not an empty pyparsing.ParseResults otherwise
      # serialization will fail.
      return None

    return value

  def _ParseFinalize(self, parser_mediator):  # pylint: disable=unused-argument
    """Finalizes parsing.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    return

  def _ParseHeader(self, parser_mediator, text_reader):  # pylint: disable=unused-argument
    """Parses a text-log file header.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Raises:
      ParseError: when the header cannot be parsed.
    """
    return

  def _ParseLines(self, parser_mediator, text_reader):
    """Parses lines of text using a pyparsing definition.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.
    """
    consecutive_line_failures = 0

    try:
      text_reader.ReadLines()
      self._current_offset = text_reader.get_offset()
    except UnicodeDecodeError as exception:
      parser_mediator.ProduceExtractionWarning((
          'unable to read and decode log line at offset {0:d} with error: '
          '{1!s}').format(self._current_offset, exception))
      return

    while text_reader.lines:
      if parser_mediator.abort:
        break

      if consecutive_line_failures > self._MAXIMUM_CONSECUTIVE_LINE_FAILURES:
        parser_mediator.ProduceExtractionWarning(
            'more than {0:d} consecutive failures to parse lines.'.format(
                self._MAXIMUM_CONSECUTIVE_LINE_FAILURES))
        break

      try:
        key, structure, _, end = self._ParseString(text_reader.lines)

      except errors.ParseError as exception:
        line = text_reader.ReadLine()
        # Pyparsing does not appear to detect single empty lines hence that
        # we ignore them here.
        if not line:
          continue

        logger.debug('unable to parse string with error: {0!s}'.format(
            exception))

        if len(line) > 80:
          line = '{0:s}...'.format(line[:77])

        parser_mediator.ProduceExtractionWarning(
            'unable to parse log line: {0:d} "{1:s}"'.format(
                text_reader.line_number, line))

        consecutive_line_failures += 1

        continue

      consecutive_line_failures = 0

      try:
        # TODO: use a callback per key.
        self._ParseRecord(parser_mediator, key, structure)

      except errors.ParseError as exception:
        parser_mediator.ProduceExtractionWarning(
            'unable to parse record: {0:s} with error: {1!s}'.format(
                key, exception))

      text_reader.SkipAhead(end)

      try:
        text_reader.ReadLines()
        self._current_offset = text_reader.get_offset()
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to read and decode log line at offset {0:d} with error: '
            '{1!s}').format(self._current_offset, exception))
        break

  @abc.abstractmethod
  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: when the structure type is unknown.
    """

  def _ParseString(self, string):
    """Parses a string for known grammar.

    Args:
      string (str): string.

    Returns:
      tuple[str, pyparsing.ParseResults, int, int]: key, parsed tokens, start
          and end offset.

    Raises:
      ParseError: when the string cannot be parsed by the grammar.
    """
    try:
      structure_generator = self._pyparsing_grammar.scanString(
          string, maxMatches=1)
      structure, start, end = next(structure_generator)

    except StopIteration:
      structure = None

    except pyparsing.ParseException as exception:
      raise errors.ParseError(exception)

    if not structure:
      raise errors.ParseError('No match found.')

    if start > 0 and '\n' in string[:start + 1]:
      raise errors.ParseError('Found a line preceeding match.')

    # Unwrap the line structure and retrieve its name (key).
    keys = list(structure.keys())
    if len(keys) != 1:
      raise errors.ParseError('Missing key of line structructure.')

    return keys[0], structure[0], start, end

  def _SetLineStructures(self, line_structures):
    """Sets the line structures.

    Args:
      line_structures ([(str, pyparsing.ParserElement)]): tuples of pyparsing
          expressions to parse a line and their names.
    """
    self._pyparsing_grammar = None

    for key, expression in line_structures:
      # Wrap the line structures in groups with a result name to build a single
      # pyparsing grammar.
      if not isinstance(expression, pyparsing.Group):
        expression = pyparsing.Group(expression).setResultsName(key)

      if not self._pyparsing_grammar:
        self._pyparsing_grammar = expression
      else:
        self._pyparsing_grammar ^= expression

    # Override Pyparsing's default replacement of tabs with spaces to
    # SkipAhead() the correct number of bytes after a match.
    self._pyparsing_grammar.parseWithTabs()

    # Override Pyparsing's whitespace characters to spaces only.
    self._pyparsing_grammar.setDefaultWhitespaceChars(' ')

  def _VerifyString(self, string):
    """Checks a string for known grammar.

    Args:
      string (str): string.

    Returns:
      pyparsing.ParseResults: parsed tokens.

    Raises:
      ParseError: when the string cannot be parsed by the grammar.
    """
    try:
      structure = self.VERIFICATION_GRAMMAR.parseString(string)
    except pyparsing.ParseException as exception:
      raise errors.ParseError(exception)

    if not structure:
      raise errors.ParseError('No match found.')

    return structure

  @abc.abstractmethod
  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """

  # pylint: disable=arguments-differ
  def Process(self, parser_mediator, file_object=None, **kwargs):
    """Extracts events from a text log file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (Optional[dfvfs.FileIO]): a file-like object.
    """
    # This will raise if unhandled keyword arguments are passed.
    super(TextPlugin, self).Process(parser_mediator)

    # Keep a reference to the parser mediator for the encoding error handler.
    self._parser_mediator = parser_mediator

    try:
      # Set the offset to the beginning of the file.
      file_object.seek(0, os.SEEK_SET)
      self._current_offset = 0

      encoding = self.ENCODING
      if not encoding:
        encoding = parser_mediator.GetCodePage()

      text_reader = text_parser.EncodedTextReader(
          file_object, encoding=encoding, encoding_errors='text_parser_handler')

      try:
        text_reader.ReadLines()
        self._current_offset = text_reader.get_offset()
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to read and decode log line at offset {0:d} with error: '
            '{1!s}').format(self._current_offset, exception))
        return

      try:
        self._ParseHeader(parser_mediator, text_reader)
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to parser header with error: {0!s}').format(exception))
        return

      self._ParseLines(parser_mediator, text_reader)

      self._ParseFinalize(parser_mediator)

    finally:
      self._parser_mediator = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the USBStor Windows Registry plugin."""

import unittest

from plaso.parsers.winreg_plugins import usbstor

from tests.parsers.winreg_plugins import test_lib


class USBStorPlugin(test_lib.RegistryPluginTestCase):
  """Tests for the USBStor Windows Registry plugin."""

  def testFilters(self):
    """Tests the FILTERS class attribute."""
    plugin = usbstor.USBStorPlugin()

    key_path = 'HKEY_LOCAL_MACHINE\\System\\ControlSet001\\Enum\\USBSTOR'
    self._AssertFiltersOnKeyPath(plugin, key_path)

    self._AssertNotFiltersOnKeyPath(plugin, 'HKEY_LOCAL_MACHINE\\Bogus')

  def testProcess(self):
    """Tests the Process function."""
    test_file_entry = self._GetTestFileEntry(['SYSTEM'])
    key_path = 'HKEY_LOCAL_MACHINE\\System\\ControlSet001\\Enum\\USBSTOR'

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = usbstor.USBStorPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'windows:registry:usbstor:instance',
        'device_last_arrival_time': None,
        'device_last_removal_time': None,
        'device_type': 'Disk',
        'display_name': 'HP v100w USB Device',
        'driver_first_installation_time': '2011-04-01T04:52:38.6860000+00:00',
        'driver_last_installation_time': '2011-04-01T04:52:38.6860000+00:00',
        'firmware_time': None,
        'key_path': (
            '{0:s}\\Disk&Ven_HP&Prod_v100w&Rev_1024\\AA951D0000007252&0'.format(
                key_path)),
        'product': 'Prod_v100w',
        'revision': 'Rev_1024',
        'vendor': 'Ven_HP'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for Santa log text parser plugin."""

import unittest

from plaso.parsers.text_plugins import santa

from tests.parsers.text_plugins import test_lib


class SantaTextPluginTest(test_lib.TextPluginTestCase):
  """Tests for Santa log text parser plugin."""

  def testProcessWithLegacyFormat(self):
    """Tests the Process function with the legacy Santa log format."""
    plugin = santa.SantaTextPlugin()
    storage_writer = self._ParseTextFileWithPlugin(['santa.log'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 191)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Execution event with quarantine URL.
    expected_event_values = {
        'data_type': 'santa:execution',
        'decision': 'ALLOW',
        'last_run_time': '2018-08-19T03:17:55.765+00:00',
        'process_hash': (
            '78b43a13b5b608fe1a5a590f5f3ff112ff16ece7befc29fc84347125f6b9ca78'),
        'process_path': '/Applications/Skype.app/Contents/MacOS/Skype',
        'quarantine_url': (
            'https://endpoint920510.azureedge.net/s4l/s4l/download/mac/'
            'Skype-8.28.0.41.dmg')}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 35)
    self.CheckEventData(event_data, expected_event_values)

    # File operation event log
    expected_event_values = {
        'action': 'WRITE',
        'data_type': 'santa:file_system_event',
        'file_path': '/Users/qwerty/newfile',
        'gid': '20',
        'group': 'staff',
        'last_written_time': '2018-08-19T04:02:47.911+00:00',
        'pid': '303',
        'ppid': '1',
        'process': 'Finder',
        'process_path': (
            '/System/Library/CoreServices/Finder.app/Contents/MacOS/Finder'),
        'uid': '501',
        'user': 'qwerty'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 142)
    self.CheckEventData(event_data, expected_event_values)

    # Test a Disk mounts event.
    expected_event_values = {
        'action': 'DISKAPPEAR',
        'appearance_time': '2018-08-19T03:17:28.982+00:00',
        'bsd_name': 'disk2s1',
        'bus': 'Virtual Interface',
        'data_type': 'santa:diskmount',
        'dmg_path': '/Users/qwerty/Downloads/Skype-8.28.0.41.dmg',
        'fs': 'hfs',
        'last_written_time': '2018-08-19T03:17:29.036+00:00',
        'model': 'Apple Disk Image',
        'mount': None,
        'serial': None,
        'volume': 'Skype'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 27)
    self.CheckEventData(event_data, expected_event_values)

  def testProcessWithCurrentFormat(self):
    """Tests the Process function with the current Santa log format."""
    plugin = santa.SantaTextPlugin()
    storage_writer = self._ParseTextFileWithPlugin(['santa2.log'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 11)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # File rename operation event log
    expected_event_values = {
        'action': 'RENAME',
        'data_type': 'santa:file_system_event',
        'file_new_path': '/private/var/db/santa/santa.log.6.gz',
        'file_path': '/private/var/db/santa/santa.log.5.gz',
        'gid': '0',
        'group': 'wheel',
        'last_written_time': '2022-02-17T16:30:05.253+00:00',
        'pid': '26150',
        'pid_version': '2228280',
        'ppid': '1',
        'process': 'newsyslog',
        'process_path': '/usr/sbin/newsyslog',
        'uid': '0',
        'user': 'root'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 3)
    self.CheckEventData(event_data, expected_event_values)

    # Process exit log
    expected_event_values = {
        'action': 'EXIT',
        'data_type': 'santa:process_exit',
        'exit_time': '2022-02-09T16:47:02.893+00:00',
        'gid': '1',
        'pid': '75780',
        'pid_version': '1765713',
        'ppid': '155',
        'uid': '0'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)

    # Link file operation event log
    expected_event_values = {
        'action': 'LINK',
        'data_type': 'santa:file_system_event',
        'file_new_path': '/private/var/db/santa/santa.log.0',
        'file_path': '/private/var/db/santa/santa.log',
        'gid': '0',
        'group': 'wheel',
        'last_written_time': '2022-02-17T16:30:05.257+00:00',
        'pid': '26150',
        'pid_version': '2228280',
        'ppid': '1',
        'process': 'newsyslog',
        'process_path': '/usr/sbin/newsyslog',
        'uid': '0',
        'user': 'root'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 4)
    self.CheckEventData(event_data, expected_event_values)

    # Execution event log
    expected_event_values = {
        'action': 'EXEC',
        'certificate_common_name': 'Software Signing',
        'certificate_hash':
            'd84db96af8c2e60ac4c851a21ec460f6f84e0235beb17d24a78712b9b021ed57',
        'data_type': 'santa:execution',
        'decision': 'ALLOW',
        'gid': '0',
        'group': 'wheel',
        'last_run_time': '2022-02-17T16:31:28.414+00:00',
        'long_reason': 'critical system binary',
        'mode': 'M',
        'pid': '26364',
        'pid_version': '2228711',
        'ppid': '1',
        'process_arguments': (
            'xpcproxy com.apple.mdworker.shared.'
            '07000000-0000-0000-0000-000000000000'),
        'process_hash':
            '25977d1584525b571fc6d19dcac1a768d0555c58777876869255de312dcfcaf3',
        'process_path': '/usr/libexec/xpcproxy',
        'reason': 'BINARY',
        'uid': '0',
        'user': 'root'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 6)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Windows EventLog providers helper."""


class WindowsEventLogProvidersHelper(object):
  """Windows EventLog providers helper."""

  def _GetNormalizedPath(self, path):
    """Retrieves a normalized variant of a path.

    Args:
      path (str): path of a message file.

    Returns:
      str: normalized path of a message file.
    """
    path_segments = path.split('\\')

    if path_segments:
      # Check if the first path segment is a drive letter or "%SystemDrive%".
      first_path_segment = path_segments[0].lower()
      if ((len(first_path_segment) == 2 and first_path_segment[1:] == ':') or
          first_path_segment == '%systemdrive%'):
        path_segments[0] = ''

    return '\\'.join(path_segments) or '\\'

  def Merge(self, first_event_log_provider, second_event_log_provider):
    """Merges the information of the second Event Log provider into the first.

    Args:
      first_event_log_provider (EventLogProvider): first Event Log provider.
      second_event_log_provider (EventLogProvider): second Event Log provider.
    """
    if not first_event_log_provider.identifier:
      first_event_log_provider.identifier = second_event_log_provider.identifier
    elif (first_event_log_provider.identifier !=
          second_event_log_provider.identifier):
      first_event_log_provider.additional_identifier = (
          second_event_log_provider.identifier)

    for log_source in second_event_log_provider.log_sources:
      log_source = log_source.lower()
      if log_source not in first_event_log_provider.log_sources:
        first_event_log_provider.log_sources.append(log_source)

    for log_type in second_event_log_provider.log_types:
      log_type = log_type.lower()
      if log_type not in first_event_log_provider.log_types:
        first_event_log_provider.log_types.append(log_type)

    for path in second_event_log_provider.category_message_files:
      path = self._GetNormalizedPath(path.lower())
      if path not in first_event_log_provider.category_message_files:
        first_event_log_provider.category_message_files.append(path)

    for path in second_event_log_provider.event_message_files:
      path = self._GetNormalizedPath(path.lower())
      if path not in first_event_log_provider.event_message_files:
        first_event_log_provider.event_message_files.append(path)

    for path in second_event_log_provider.parameter_message_files:
      path = self._GetNormalizedPath(path.lower())
      if path not in first_event_log_provider.parameter_message_files:
        first_event_log_provider.parameter_message_files.append(path)

  def NormalizeMessageFiles(self, event_log_provider):
    """Normalizes the message files.

    Args:
      event_log_provider (EventLogProvider): Event Log provider.
    """
    event_log_provider.category_message_files = [
        self._GetNormalizedPath(path.lower())
        for path in event_log_provider.category_message_files]

    event_log_provider.event_message_files = [
        self._GetNormalizedPath(path.lower())
        for path in event_log_provider.event_message_files]

    event_log_provider.parameter_message_files = [
        self._GetNormalizedPath(path.lower())
        for path in event_log_provider.parameter_message_files]
